using System;
using System.IO;
using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	public class DERInputStream : FilterInputStream 
	{
		public DERInputStream(
			Stream ins) : base(ins) {}

		protected virtual int readLength()
			//throws IOException
		{
			int length = read();
			if (length < 0)
			{
				throw new IOException("EOF found when length expected");
			}

			if (length == 0x80)
			{
				return -1;      // indefinite-length encoding
			}

			if (length > 127)
			{
				int size = length & 0x7f;

				length = 0;
				for (int i = 0; i < size; i++)
				{
					int next = read();

					if (next < 0)
					{
						throw new IOException("EOF found reading length");
					}

					length = (length << 8) + next;
				}
			}

			return length;
		}

		protected virtual void readFully(
			byte[]  bytes)
			//throws IOException
		{
			int     left = bytes.Length;

			if (left == 0)
			{
				return;
			}

			while ((left -= read(bytes, bytes.Length - left, left)) != 0)
			{
				;
			}
		}

		/**
		* build an object given its tag and a byte stream to construct it
		* from.
		*/
		protected virtual DERObject buildObject(
			int	    tag,
			byte[]	bytes)
			//throws IOException
		{
			if (tag == DERTags.NULL) return null;   
			if (tag == (DERTags.SEQUENCE | DERTags.CONSTRUCTED))
			{
				MemoryStream    bIn = new MemoryStream(bytes);
				BERInputStream          dIn = new BERInputStream(bIn);
				DERConstructedSequence  seq = new DERConstructedSequence();

				try
				{
					for (;;)
					{
						DERObject   obj = dIn.readObject();

						seq.addObject(obj);
					}
				}
				catch (EOFException ex)
				{
					ex.ToString();
					return seq;
				}
			}
			if (tag == (DERTags.SET | DERTags.CONSTRUCTED))
			{
				MemoryStream bIn = new MemoryStream(bytes);
				BERInputStream dIn = new BERInputStream(bIn);

				ASN1EncodableVector    v = new ASN1EncodableVector();

				try
				{
					for (;;)
					{
						DERObject   obj = dIn.readObject();

						v.add(obj);
					}
				}
				catch (EOFException ex)
				{
					ex.ToString();
					return new DERConstructedSet(v);
				}
			}
			if (tag == DERTags.BOOLEAN) return new DERBoolean(bytes);
			if (tag == DERTags.INTEGER) return new DERInteger(bytes);
			if (tag == DERTags.ENUMERATED) return new DEREnumerated(bytes);
			if (tag == DERTags.OBJECT_IDENTIFIER) return new DERObjectIdentifier(bytes);
			if (tag == DERTags.BIT_STRING)
			{
				int     padBits = bytes[0];
				byte[]  data = new byte[bytes.Length - 1];

				Array.Copy(bytes, 1, data, 0, bytes.Length - 1);

				return new DERBitString(data, padBits);
			}
			if (tag == DERTags.UTF8_STRING) return new DERUTF8String(bytes);
			if (tag == DERTags.PRINTABLE_STRING) return new DERPrintableString(bytes);
			if (tag == DERTags.IA5_STRING) return new DERIA5String(bytes);
			if (tag == DERTags.T61_STRING) return new DERT61String(bytes);
			if (tag == DERTags.VISIBLE_STRING) return new DERVisibleString(bytes);
			if (tag == DERTags.UNIVERSAL_STRING) return new DERUniversalString(bytes);
			if (tag == DERTags.BMP_STRING) return new DERBMPString(bytes);
			if (tag == DERTags.OCTET_STRING) return new DEROctetString(bytes);
			if (tag == DERTags.UTC_TIME) return new DERUTCTime(bytes);
			if (tag == DERTags.GENERALIZED_TIME) return new DERGeneralizedTime(bytes);

			//default:
				//
				// with tagged object tag number is bottom 5 bits
				//
			if ((tag & DERTags.TAGGED) != 0)  
			{
				if ((tag & 0x1f) == 0x1f)
				{
					throw new IOException("unsupported high tag encountered");
				}

				if (bytes.Length == 0)        // empty tag!
				{
					if ((tag & DERTags.CONSTRUCTED) == 0)
					{
						return new DERTaggedObject(false, tag & 0x1f, new DERNull());
					}
					else
					{
						return new DERTaggedObject(false, tag & 0x1f, new DERConstructedSequence());
					}
				}

				//
				// simple type - implicit... return an octet string
				//
				if ((tag & DERTags.CONSTRUCTED) == 0)
				{
					return new DERTaggedObject(false, tag & 0x1f, new DEROctetString(bytes));
				}

				MemoryStream bIn = new MemoryStream(bytes);
				BERInputStream dIn = new BERInputStream(bIn);

				DEREncodable dObj = dIn.readObject();

				//
				// explicitly tagged (probably!) - if it isn't we'd have to
				// tell from the context
				//
				if (dIn.available() == 0)
				{
					return new DERTaggedObject(tag & 0x1f, dObj);
				}

				//
				// another implicit object, we'll create a sequence...
				//
				DERConstructedSequence  seq = new DERConstructedSequence();

				seq.addObject(dObj);

				try
				{
					for (;;)
					{
						dObj = dIn.readObject();

						seq.addObject(dObj);
					}
				}
				catch (EOFException ex)
				{
					ex.ToString();
					// ignore --
				}

				return new DERTaggedObject(false, tag & 0x1f, seq);
			}
			else 
			{
				return new DERUnknownTag(tag, bytes);
			}
		}

		public virtual DERObject readObject()
			//throws IOException
		{
			int tag = read();
			if (tag == -1)
			{
				throw new EOFException();
			}

			int     length = readLength();
			byte[]  bytes = new byte[length];

			readFully(bytes);

			return buildObject(tag, bytes);
		}
	}

}