using System;
using System.Collections;
using System.Text;
using org.bouncycastle.io;
using org.bouncycastle.asn1;

namespace org.bouncycastle.asn1.x509
{
	public class X509Name: DEREncodable
	{
		/**
		* country code - StringType(SIZE(2))
		*/
		public static readonly DERObjectIdentifier C = new DERObjectIdentifier("2.5.4.6");

		/**
		* organization - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier O = new DERObjectIdentifier("2.5.4.10");

		/**
		* organizational unit name - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier OU = new DERObjectIdentifier("2.5.4.11");

		/**
		* Title
		*/
		public static readonly DERObjectIdentifier T = new DERObjectIdentifier("2.5.4.12");

		/**
		* common name - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier CN = new DERObjectIdentifier("2.5.4.3");

		/**
		* device serial number name - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier SN = new DERObjectIdentifier("2.5.4.5");

		/**
		* locality name - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier L = new DERObjectIdentifier("2.5.4.7");

		/**
		* state, or province name - StringType(SIZE(1..64))
		*/
		public static readonly DERObjectIdentifier ST = new DERObjectIdentifier("2.5.4.8");

		/**
		* Naming attributes of type X520name
		*/
		public static readonly DERObjectIdentifier SURNAME = new DERObjectIdentifier("2.5.4.4");
		public static readonly DERObjectIdentifier GIVENNAME = new DERObjectIdentifier("2.5.4.42");
		public static readonly DERObjectIdentifier INITIALS = new DERObjectIdentifier("2.5.4.43");
		public static readonly DERObjectIdentifier GENERATION = new DERObjectIdentifier("2.5.4.44");


		/**
		* email address (RSA PKCS#9 extension) - IA5String
		* <p>
		* note: if you're trying to be ultra orthodox, don't use this! It shouldn't be in here.
		*/
		public static readonly DERObjectIdentifier EmailAddress = new DERObjectIdentifier("1.2.840.113549.1.9.1");
		
		/**
		* email address in Verisign certificates
		*/
		public static readonly DERObjectIdentifier E = EmailAddress;
		
		/*
		* others...
		*/
		public static readonly DERObjectIdentifier DC = new DERObjectIdentifier("0.9.2342.19200300.100.1.25");

		/**
		* LDAP User id.
		*/
		public static readonly DERObjectIdentifier UID = new DERObjectIdentifier("0.9.2342.19200300.100.1.1");

		/**
		* look up table translating OID values into their common symbols - this static is scheduled for deletion
		*/
		public static Hashtable OIDLookUp = new Hashtable();

		/**
		* determines whether or not strings should be processed and printed
		* from back to front.
		*/
		public static bool DefaultReverse = false;

		/**
		* default look up table translating OID values into their common symbols following
		* the convention in RFC 2253 with a few extras
		*/
		public static Hashtable DefaultSymbols = OIDLookUp;

		/**
		* look up table translating OID values into their common symbols following the convention in RFC 2253
		* with a few extras
		*/
		public static Hashtable RFC2253Symbols = new Hashtable();

		/**
		* look up table translating string values into their OIDS -
		* this static is scheduled for deletion
		*/
		public static Hashtable SymbolLookUp = new Hashtable();

		/**
		* look up table translating common symbols into their OIDS.
		*/
		public static Hashtable DefaultLookUp = SymbolLookUp;

		static X509Name()
		{
			DefaultSymbols[C] = "C";
			DefaultSymbols[O] = "O";
			DefaultSymbols[T] = "T";
			DefaultSymbols[OU] = "OU";
			DefaultSymbols[CN] = "CN";
			DefaultSymbols[L] = "L";
			DefaultSymbols[ST] = "ST";
			DefaultSymbols[SN] = "SN";
			DefaultSymbols[EmailAddress] = "E";
			DefaultSymbols[DC] = "DC";
			DefaultSymbols[UID] = "UID";
			DefaultSymbols[SURNAME] = "SURNAME";
			DefaultSymbols[GIVENNAME] = "GIVENNAME";
			DefaultSymbols[INITIALS] = "INITIALS";
			DefaultSymbols[GENERATION] = "GENERATION";

			RFC2253Symbols[C] = "C";
			RFC2253Symbols[O] = "O";
			RFC2253Symbols[T] = "T";
			RFC2253Symbols[OU] = "OU";
			RFC2253Symbols[CN] = "CN";
			RFC2253Symbols[L] = "L";
			RFC2253Symbols[ST] = "ST";
			RFC2253Symbols[SN] = "SN";
			RFC2253Symbols[EmailAddress] = "EMAILADDRESS";
			RFC2253Symbols[DC] = "DC";
			RFC2253Symbols[UID] = "UID";
			RFC2253Symbols[SURNAME] = "SURNAME";
			RFC2253Symbols[GIVENNAME] = "GIVENNAME";
			RFC2253Symbols[INITIALS] = "INITIALS";
			RFC2253Symbols[GENERATION] = "GENERATION";

			DefaultLookUp["c"] = C;
			DefaultLookUp["o"] = O;
			DefaultLookUp["t"] = T;
			DefaultLookUp["ou"] = OU;
			DefaultLookUp["cn"] = CN;
			DefaultLookUp["l"] = L;
			DefaultLookUp["st"] = ST;
			DefaultLookUp["sn"] = SN;
			DefaultLookUp["emailaddress"] = E;
			DefaultLookUp["dc"] = DC;
			DefaultLookUp["e"] = E;
			DefaultLookUp["uid"] = UID;
			DefaultLookUp["surname"] = SURNAME;
			DefaultLookUp["givenname"] = GIVENNAME;
			DefaultLookUp["initials"] = INITIALS;
			DefaultLookUp["generation"] = GENERATION;
		}

		private ArrayList                  ordering = new ArrayList();
		private ArrayList                  values = new ArrayList();
		private ASN1Sequence            seq;

		public static X509Name getInstance(
			ASN1TaggedObject obj,
			bool          ArrayList)
		{
			return getInstance(ASN1Sequence.getInstance(obj, ArrayList));
		}

		public static X509Name getInstance(
			Object  obj)
		{
			
			if ((obj == null) || (typeof(X509Name).IsInstanceOfType(obj)))
			{
				return (X509Name)obj;
			}
			else if (typeof(ASN1Sequence).IsInstanceOfType(obj))
			{
				return new X509Name((ASN1Sequence)obj);
			}

			throw new ArgumentException("unknown object in factory");
		}

		/**
		* Constructor from ASN1Sequence
		*
		* the principal will be a list of constructed sets, each containing an (OID, String) pair.
		*/
		public X509Name(
			ASN1Sequence  seq)
		{
			this.seq = seq;

			IEnumerator e = seq.getObjects();

			while (e.MoveNext())
			{
				ASN1Set         set = (ASN1Set)e.Current;
				ASN1Sequence    s = (ASN1Sequence)set.getObjectAt(0);

				ordering.Add(s.getObjectAt(0));
				values.Add(((DERString)s.getObjectAt(1)).getString());
			}
		}

		/**
		* constructor from a table of attributes.
		* <p>
		* it's is assumed the table contains OID/String pairs, and the contents
		* of the table are copied into an internal table as part of the 
		* construction process.
		* <p>
		* <b>Note:</b> if the name you are trying to generate should be
		* following a specific ordering, you should use the constructor
		* with the ordering specified below.
		*/
		public X509Name(
			Hashtable  attributes) : this(null, attributes) {}

		/**
		* constructor from a table of attributes with ordering.
		* <p>
		* it's is assumed the table contains OID/String pairs, and the contents
		* of the table are copied into an internal table as part of the 
		* construction process. The ordering vector should contain the OIDs
		* in the order they are meant to be encoded or printed in ToString.
		*/
		public X509Name(
			ArrayList      ordering,
			Hashtable   attributes)
		{
			if (ordering != null)
			{
				for (int i = 0; i != ordering.Count; i++)
				{
					this.ordering.Add(ordering[i]);
				}
			}
			else
			{
				IEnumerator     e = attributes.Keys.GetEnumerator();

				while (e.MoveNext())
				{
					this.ordering.Add(e.Current);
				}
			}

			for (int i = 0; i != this.ordering.Count; i++)
			{
				DERObjectIdentifier     oid = (DERObjectIdentifier)this.ordering[i];

				if (attributes[oid] == null)
				{
					throw new ArgumentException("No attribute for object id - " + oid.getId() + " - passed to distinguished name");
				}

				this.values.Add(attributes[oid]); // copy the hash table
			}
		}

		/**
		* takes two vectors one of the oids and the other of the values.
		*/
		public X509Name(
			ArrayList  oids,
			ArrayList  values)
		{
			if (oids.Count != values.Count)
			{
				throw new ArgumentException("oids vector must be same length as values.");
			}

			for (int i = 0; i < oids.Count; i++)
			{
				this.ordering.Add(oids[i]);
				this.values.Add(values[i]);
			}
		}

		/**
		* takes an X509 dir name as a string of the format "C=AU, ST=Victoria", or
		* some such, converting it into an ordered set of name attributes.
		*/
		public X509Name(
			String  dirName): this(DefaultReverse, DefaultLookUp, dirName) {}

		/**
		* takes an X509 dir name as a string of the format "C=AU, ST=Victoria", or
		* some such, converting it into an ordered set of name attributes. If reverse
		* is true, create the encoded version of the sequence starting from the
		* last element in the string.
		*/
		public X509Name(
			bool reverse,
			String  dirName): this(reverse, DefaultLookUp, dirName) {}

		/**
		* takes an X509 dir name as a string of the format "C=AU, ST=Victoria", or
		* some such, converting it into an ordered set of name attributes. lookUp 
		* should provide a table of lookups, indexed by lowercase only strings and
		* yielding a DERObjectIdentifier, other than that OID. and numeric oids
		* will be processed automatically.
		* <p>
		* If reverse is true, create the encoded version of the sequence starting
		* from the last element in the string.
		*/
		public X509Name(
			bool     reverse,
			Hashtable   lookUp,
			String      dirName)
		{
			X509NameTokenizer   nTok = new X509NameTokenizer(dirName);

			while (nTok.hasMoreTokens())
			{
				String  token = nTok.nextToken();
				int     index = token.IndexOf('=');

				if (index == -1)
				{
					throw new ArgumentException("badly formated directory string");
				}

				String              name = token.Substring(0, index);
				String              value = token.Substring(index + 1);
				DERObjectIdentifier oid = null;

				if (name.ToUpper().StartsWith("OID."))
				{
					oid = new DERObjectIdentifier(name.Substring(4));
				}
				else if (name[0] >= '0' && name[0] <= '9')
				{
					oid = new DERObjectIdentifier(name);
				}
				else
				{
					oid = (DERObjectIdentifier)lookUp[name.ToLower()];
					if (oid == null)
					{
						throw new ArgumentException("Unknown object id - " + name + " - passed to distinguished name");
					}
				}

				this.ordering.Add(oid);
				this.values.Add(value);
			}

			if (reverse)
			{
				ArrayList  o = new ArrayList();
				ArrayList  v = new ArrayList();

				for (int i = this.ordering.Count - 1; i >= 0; i--)
				{
					o.Add(this.ordering[i]);
					v.Add(this.values[i]);
				}

				this.ordering = o;
				this.values = v;
			}
		}

		/**
		* return a vector of the oids in the name, in the order they were found.
		*/
		public ArrayList getOIDs()
		{
			ArrayList  v = new ArrayList();

			for (int i = 0; i != ordering.Count; i++)
			{
				v.Add(ordering[i]);
			}

			return v;
		}

		/**
		* return a vector of the values found in the name, in the order they
		* were found.
		*/
		public ArrayList getValues()
		{
			ArrayList  v = new ArrayList();

			for (int i = 0; i != values.Count; i++)
			{
				v.Add(values[i]);
			}

			return v;
		}

		/**
		* return false if we have characters out of the range of a printable
		* string, true otherwise.
		*/
		private bool canBePrintable(
			String  str)
		{
			for (int i = str.Length - 1; i >= 0; i--)
			{
				if (str[i] > 0x007f)
				{
					return false;
				}
			}

			return true;
		}

		public DERObject getDERObject()
		{
			if (seq == null)
			{
				ASN1EncodableVector  vec = new ASN1EncodableVector();

				for (int i = 0; i != ordering.Count; i++)
				{
					ASN1EncodableVector     v = new ASN1EncodableVector();
					DERObjectIdentifier     oid = (DERObjectIdentifier)ordering[i];

					v.add(oid);

					String  str = (String)values[i];

					if (oid.Equals(EmailAddress))
					{
						v.add(new DERIA5String(str));
					}
					else
					{
						if (canBePrintable(str))
						{
							v.add(new DERPrintableString(str));
						}
						else
						{
							v.add(new DERUTF8String(str));
						}
					}

					vec.add(new DERSet(new DERSequence(v)));
				}

				seq = new DERSequence(vec);
			}

			return seq;
		}

		/**
		* @param inOrder if true the order of both X509 names must be the same,
		* as well as the values associated with each element.
		*/
		public virtual bool Equals(Object _obj, bool inOrder) 
		{
			if (_obj == this)
			{
				return true;
			}

			if (!inOrder)
			{
				return this.Equals(_obj);
			}

			if (_obj == null || !(typeof(X509Name).IsInstanceOfType(_obj)))
			{
				return false;
			}
	        
			X509Name _oxn          = (X509Name)_obj;
			int      _orderingSize = ordering.Count;

			if (_orderingSize != _oxn.ordering.Count) 
			{
				return false;
			}
			
			for(int i = 0; i < _orderingSize; i++) 
			{
				String  _oid   = ((DERObjectIdentifier)ordering[i]).getId();
				String  _val   = (String)values[i];
				
				String _oOID = ((DERObjectIdentifier)_oxn.ordering[i]).getId();
				String _oVal = (String)_oxn.values[i];

				if (_oid.Equals(_oOID))
				{
					_val = _val.Trim().ToLower();
					_oVal = _oVal.Trim().ToLower();
					if (_val.Equals(_oVal))
					{
						continue;
					}
					else
					{
						StringBuilder    v1 = new StringBuilder();
						StringBuilder    v2 = new StringBuilder();

						if (_val.Length != 0)
						{
							char    c1 = _val[0];

							v1.Append(c1);

							for (int k = 1; k < _val.Length; k++)
							{
								char    c2 = _val[k];
								if (!(c1 == ' ' && c2 == ' '))
								{
									v1.Append(c2);
								}
								c1 = c2;
							}
						}

						if (_oVal.Length != 0)
						{
							char    c1 = _oVal[0];

							v2.Append(c1);

							for (int k = 1; k < _oVal.Length; k++)
							{
								char    c2 = _oVal[k];
								if (!(c1 == ' ' && c2 == ' '))
								{
									v2.Append(c2);
								}
								c1 = c2;
							}
						}

						if (!v1.ToString().Equals(v2.ToString()))
						{
							return false;
						}
					}
				}
			}

			return true;
		}

		/**
		* test for equality - note: case is ignored.
		*/
		public override bool Equals(Object _obj) 
		{
			if (_obj == this)
			{
				return true;
			}

			if (_obj == null || !(typeof(X509Name).IsInstanceOfType(_obj)))
			{
				return false;
			}
	        
			X509Name _oxn          = (X509Name)_obj;
			int      _orderingSize = ordering.Count;

			if (_orderingSize != _oxn.ordering.Count) 
			{
				return false;
			}
			
			bool[] _indexes = new bool[_orderingSize];

			for(int i = 0; i < _orderingSize; i++) 
			{
				bool _found = false;
				String  _oid   = ((DERObjectIdentifier)ordering[i]).getId();
				String  _val   = (String)values[i];
				
				for(int j = 0; j < _orderingSize; j++) 
				{
					if(_indexes[j] == true)
					{
						continue;
					}
					
					String _oOID = ((DERObjectIdentifier)_oxn.ordering[j]).getId();
					String _oVal = (String)_oxn.values[j];

					if (_oid.Equals(_oOID))
					{
						_val = _val.Trim().ToLower();
						_oVal = _oVal.Trim().ToLower();
						if (_val.Equals(_oVal))
						{
							_indexes[j] = true;
							_found      = true;
							break;
						}
						else
						{
							StringBuilder    v1 = new StringBuilder();
							StringBuilder    v2 = new StringBuilder();

							if (_val.Length != 0)
							{
								char    c1 = _val[0];

								v1.Append(c1);

								for (int k = 1; k < _val.Length; k++)
								{
									char    c2 = _val[k];
									if (!(c1 == ' ' && c2 == ' '))
									{
										v1.Append(c2);
									}
									c1 = c2;
								}
							}

							if (_oVal.Length != 0)
							{
								char    c1 = _oVal[0];

								v2.Append(c1);

								for (int k = 1; k < _oVal.Length; k++)
								{
									char    c2 = _oVal[k];
									if (!(c1 == ' ' && c2 == ' '))
									{
										v2.Append(c2);
									}
									c1 = c2;
								}
							}

							if (v1.ToString().Equals(v2.ToString()))
							{
								_indexes[j] = true;
								_found      = true;
								break;
							}
						}
					}
				}

				if(!_found)
				{
					return false;
				}
			}
			
			return true;
		}
		
		public override  int GetHashCode()
		{
			ASN1Sequence  seq = (ASN1Sequence)this.getDERObject();
			IEnumerator   e = seq.getObjects();
			int           hashCode = 0;

			while (e.MoveNext())
			{
				hashCode ^= e.Current.GetHashCode();
			}

			return hashCode;
		}

		private void appendValue(
			StringBuilder        buf,
			Hashtable           oidSymbols,
			DERObjectIdentifier oid,
			String              value)
		{
			String  sym = (String)oidSymbols[oid];

			if (sym != null)
			{
				buf.Append(sym);
			}
			else
			{
				buf.Append(((DERObjectIdentifier)oid).getId());
			}

			buf.Append("=");

			int     index = buf.Length;

			buf.Append(value);

			int     end = buf.Length;

			while (index != end)
			{
				if ((buf[index] == ',')
				|| (buf[index] == '"')
				|| (buf[index] == '\\')
				|| (buf[index] == '+')
				|| (buf[index] == '<')
				|| (buf[index] == '>')
				|| (buf[index] == ';'))
				{
					buf.Insert(index, "\\");
					index++;
					end++;
				}

				index++;
			}
		}

		/**
		* convert the structure to a string - if reverse is true the
		* oids and values are listed out starting with the last element
		* in the sequence (ala RFC 2253), otherwise the string will begin
		* with the first element of the structure. If no string definition
		* for the oid is found in oidSymbols the string value of the oid is
		* added. Two standard symbol tables are provided DefaultSymbols, and
		* RFC2253Symbols as part of this class.
		*
		* @param reverse if true start at the end of the sequence and work back.
		* @param oidSymbols look up table strings for oids.
		*/
		public String ToString(
			bool     reverse,
			Hashtable   oidSymbols)
		{
			StringBuilder            buf = new StringBuilder();
			bool                 first = true;

			if (reverse)
			{
				for (int i = ordering.Count - 1; i >= 0; i--)
				{
					if (first)
					{
						first = false;
					}
					else
					{
						buf.Append(",");
					}

					appendValue(buf, oidSymbols, 
								(DERObjectIdentifier)ordering[i],
								(String)values[i]);
				}
			}
			else
			{
				for (int i = 0; i < ordering.Count; i++)
				{
					if (first)
					{
						first = false;
					}
					else
					{
						buf.Append(",");
					}

					appendValue(buf, oidSymbols, 
								(DERObjectIdentifier)ordering[i],
								(String)values[i]);
				}
			}

			return buf.ToString();
		}

		public override  String ToString()
		{
			return ToString(DefaultReverse, DefaultSymbols);
		}
	}


}